/****************************************************************************/
/* |\   ======   /| **              IDRepl              ** |\   ======   /| */
/* |__\========/__| **       MasarLabs ID-Replacer      ** |__\========/__| */
/* |\            /| ** -------------------------------- ** |\            /| */
/* |  \        /  | **  Maurizio Sartori (c) 1998-2000  ** |  \        /  | */
/* |    \    /    | ** -------------------------------- ** |    \    /    | */
/* |    /=\/=\    | **            Written  by           ** |    /=\/=\    | */
/* |  /========\  | **     Maurizio Sartori 'masar'     ** |  /========\  | */
/* |/   ======   \| **   e-mail:  masar@MasarLabs.com   ** |/   ======   \| */
/****************************************************************************/
/*              This program comes with ABSOLUTELY NO WARRANTY              */
/*            for details refer to the GNU General Public License           */
/****************************************************************************/
/*\F $Id: adb.cpp,v 1.5 2000/12/17 18:36:05 cvs Exp $
*****************************************************************************/

static char Psz_RCS [] =
   "$Id: adb.cpp,v 1.5 2000/12/17 18:36:05 cvs Exp $";

/****************************************************************************/

#include "version.hpp"
#include "idrepl.hpp"
#include "verbose.hpp"
#include "adb.hpp"

#define BUFFER_SIZE (1024)
#define FIELD_SIZE  (64)

/****************************************************************************/
/** Constructor
***
*\C Parameters:
*\P   isIStream   IN - Stream to read from
*\P   chFieldSep  IN - Field Separator. If set to '\n' lines are returned
*\P   pcszDefault IN - Default empty string
*****************************************************************************/
ASCII_DB::ASCII_DB (istream & isIStream,
                    CHAR chFieldSep,
                    PCSZ pcszDefault) :
   psz_Buff (NULL),
   psz_BuffPos (NULL),
   psz_BuffEnd (NULL),
   i_BuffSize (0),
   apcsz_Fields (NULL),
   i_FieldsSize (0),
   i_FieldsUsed (0),
   is_IStream (isIStream),
   l_StreamPos (0),
   ch_FieldSep (chFieldSep),
   pcsz_Default (pcszDefault)
   {
   *Psz_RCS = *Psz_RCS;                 // To fake compiler warnings

   psz_Buff = new CHAR [BUFFER_SIZE + BUFFER_SIZE + 1];
   if (psz_Buff)
      {
      psz_BuffPos = psz_Buff + BUFFER_SIZE;
      psz_BuffEnd = psz_BuffPos;
      i_BuffSize  = BUFFER_SIZE;
      }

   apcsz_Fields = new PCSZ[FIELD_SIZE];
   if (apcsz_Fields)
      {
      i_FieldsSize = FIELD_SIZE;
      i_FieldsUsed = 0;
      }
   }

/****************************************************************************/
/** Clear all Fileds
***
*\C Returns:
*\R   0 == OK
*\N   1 == ERROR
*****************************************************************************/
inline APIRET ASCII_DB::ResetField (VOID)
   {
   if (apcsz_Fields == NULL)
      {
      return (1);
      }

   i_FieldsUsed = 0;

   return (0);
   }

/****************************************************************************/
/** Reallocate Fields array when array is full
***
*\C Returns:
*\R   0 == OK
*\N   1 == ERROR
*****************************************************************************/
inline APIRET ASCII_DB::ReallocField (VOID)
   {
   PCSZ * apcszTmpFields = new PCSZ [i_FieldsSize + i_FieldsSize];

   if (apcszTmpFields == NULL)
      {
      return (1);
      }

   for (INT i = 0; i < i_FieldsSize; ++i)
      {
      apcszTmpFields[i] = apcsz_Fields[i];
      }

   delete [] apcsz_Fields;

   apcsz_Fields = apcszTmpFields;
   i_FieldsSize += i_FieldsSize;

   return (0);
   }

/****************************************************************************/
/** Add a pointer to the Fields array
***
*\C Parameters:
*\P   pcszField   IN - Pointer to add
***
*\C Returns:
*\R   0 == OK
*\N   1 == ERROR
*****************************************************************************/
inline APIRET ASCII_DB::AddField (PCSZ pcszField)
   {
   if ((i_FieldsUsed == i_FieldsSize) &&
       (ReallocField () != 0))
      {
      return (1);
      }

   apcsz_Fields[i_FieldsUsed++] = pcszField;

   return (0);
   }

/****************************************************************************/
/** Reacalc Fields offset when buffer is moved
***
*\C Parameters:
*\P   pcszNewField0   IN - New pointer of Field 0
***
*\C Returns:
*\R   0 == OK
*\N   1 == ERROR
*****************************************************************************/
inline APIRET ASCII_DB::RecalcField (PCSZ pcszNewField0)
   {
   if ((apcsz_Fields == NULL) ||
       (pcszNewField0 == NULL))
      {
      return (1);
      }

   PCSZ pcszOldField0 = apcsz_Fields[0];

   for (INT i = 0; i < i_FieldsUsed; ++i)
      {
      apcsz_Fields[i] = pcszNewField0 + (apcsz_Fields[i] - pcszOldField0);
      }

   return (0);
   }

/****************************************************************************/
/** Read next line of data and parse it
***
*\C Returns:
*\R   0 == OK
*\N   1 == ERROR
*****************************************************************************/
APIRET ASCII_DB::Next (VOID)
   {
   APIRET apiRet = 1;
   PSZ pszPos;

   if (i_FieldsUsed > 0)
      {
      l_StreamPos += psz_BuffPos - apcsz_Fields[0];
      }

   if ((psz_BuffPos == NULL) ||
       ((apiRet = ResetField ()) != 0) ||
       ((apiRet = AddField (psz_BuffPos)) != 0))
      {
      return (apiRet);
      }

   if ((psz_BuffPos == psz_BuffEnd) &&
       ((apiRet = ReadBuffer ()) != 0))
      {
      ResetField ();
      return (apiRet);
      }

   apiRet = 0;
   pszPos = psz_BuffPos;
   while (*pszPos != '\n')
      {
      if (*pszPos == ch_FieldSep)
         {
         *pszPos = 0;

         if ((apiRet = AddField (pszPos + 1)) != 0)
            {
            return (apiRet);
            }
         }

      ++pszPos;
      if (pszPos == psz_BuffEnd)
         {
         apiRet = ReadBuffer ();

         if (apiRet != 0)
            {
            if (apiRet == 2)
               {
               break;
               }
            else
               {
               return (apiRet);
               }
            }

         pszPos = psz_BuffPos;
         }
      }

   if (*(pszPos - 1) == '\r')
      {
      *(pszPos - 1) = 0;
      }

   *pszPos++ = 0;

   psz_BuffPos = (pszPos <= psz_BuffEnd) ? pszPos : psz_BuffEnd;

   return (apiRet);
   }

/****************************************************************************/
/** Read the new buffer and reallocate it if needed
***
*\C Returns:
*\R   0 == OK
*\N   1 == ERROR
*\N   2 == EOF
*\N   3 == OUT_OF_MEMORY
*****************************************************************************/
APIRET ASCII_DB::ReadBuffer (VOID)
   {
   APIRET apiRet;
   PSZ pszNewBuff;
   PSZ pszNewField;

   psz_BuffPos = psz_BuffEnd;

   if (is_IStream.eof ())
      {
      return (2);
      }

   if (is_IStream.fail ())
      {
      return (1);
      }

   if (i_FieldsUsed > 0)
      {
      if (apcsz_Fields[0] < (psz_Buff + i_BuffSize))
        {
        pszNewBuff = new CHAR [4 * i_BuffSize + 1 ];
        if (pszNewBuff == NULL)
           {
           return (3);
           }

        i_BuffSize += i_BuffSize;

        pszNewField = pszNewBuff + i_BuffSize - (psz_BuffEnd - apcsz_Fields[0]);

        memcpy (pszNewField, apcsz_Fields[0], psz_BuffEnd - apcsz_Fields[0]);
        if ((apiRet = RecalcField (pszNewField)) != 0)
           {
           return (apiRet);
           }

        delete [] psz_Buff;
        psz_Buff = pszNewBuff;
        }
     else
        {
        pszNewField = psz_Buff + i_BuffSize - (psz_BuffEnd - apcsz_Fields[0]);

        memcpy (pszNewField, apcsz_Fields[0], psz_BuffEnd - apcsz_Fields[0]);

        if ((apiRet = RecalcField (pszNewField)) != 0)
           {
           return (apiRet);
           }
        }
     }

   psz_BuffPos = psz_Buff + i_BuffSize;
   is_IStream.read (psz_BuffPos, i_BuffSize);
   psz_BuffEnd = psz_BuffPos + is_IStream.gcount ();

   if (is_IStream.gcount () == 0)
      {
      return (2);
      }

   return (0);
   }

/****************************************************************************/
/** Copy current record to the ASCII_DB_RECORD object
***
*\C Parameters:
*\P   mclAdbRecord OUT - Where to copy the record
***
*\C Returns:
*\R   0 == OK
*\N   1 == ERROR
*****************************************************************************/
APIRET ASCII_DB::CopyRecord (ASCII_DB_RECORD & adbRecord)
   {
   if ((psz_Buff == NULL) || (apcsz_Fields == NULL))
      {
      return (1);
      }

   if (i_FieldsUsed == 0)
      {
      adbRecord.i_FieldsUsed = 0;
      return (0);
      }

   if ((adbRecord.psz_Buff == NULL) ||
       (adbRecord.i_BuffSize < (psz_BuffPos - apcsz_Fields[0])))
      {
      adbRecord.psz_Buff = new CHAR [psz_BuffPos - apcsz_Fields[0]];
      if (adbRecord.psz_Buff == NULL)
         {
         adbRecord.i_BuffSize = 0;
         adbRecord.i_FieldsUsed = 0;
         return (1);
         }
      adbRecord.i_BuffSize = psz_BuffPos - apcsz_Fields[0];
      }

   if ((adbRecord.apcsz_Fields == NULL) ||
       (adbRecord.i_FieldsSize < (i_FieldsUsed + 1)))
      {
      adbRecord.apcsz_Fields = new PCSZ [FIELD_SIZE];
      if (adbRecord.apcsz_Fields == NULL)
         {
         adbRecord.i_FieldsSize = 0;
         adbRecord.i_FieldsUsed = 0;
         return (1);
         }
      adbRecord.i_FieldsSize = i_FieldsUsed + 1;
      }

   memcpy (adbRecord.psz_Buff, apcsz_Fields[0],
           psz_BuffPos - apcsz_Fields[0]);

   for (INT i = 0; i < i_FieldsUsed; ++i)
      {
      adbRecord.apcsz_Fields[i] =
         adbRecord.psz_Buff + (apcsz_Fields[i] - apcsz_Fields[0]);
      }
   adbRecord.i_FieldsUsed = i_FieldsUsed;

   return (0);
   }

/****************************************************************************/
/** Constructor
***
*\C Parameters:
*\P   pcszDefault IN - Default empty string
*****************************************************************************/
ASCII_DB_RECORD::ASCII_DB_RECORD (PCSZ pcszDefault) :
   psz_Buff (NULL),
   i_BuffSize (0),
   apcsz_Fields (NULL),
   i_FieldsSize (0),
   i_FieldsUsed (0),
   pcsz_Default (pcszDefault)
   {
   }

/****************************************************************************/
/** Destructor
***
*\C Parameters:
*\P   none
*****************************************************************************/
ASCII_DB_RECORD::~ASCII_DB_RECORD (VOID)
   {
   if (psz_Buff)
      {
      delete [] psz_Buff;
      }
   if (apcsz_Fields)
      {
      delete [] apcsz_Fields;
      }
   }

/****************************************************************************/
/** Load a file into the HashMap.
***
*** File syntax is:
***   ...
***
*\C Parameters:
*\P   isIStream    IN - Stream to read from
*\P   pcszInFile   IN - Current stream name
*\P   hashMap     OUT - Where to put the data
***
*\C Returns:
*\R   0 == OK
*\N   1 == ERROR
*****************************************************************************/
APIRET LoadList (istream & isIStream, PCSZ pcszInFile,
                 HASH_MAP_STR_STR & hashMap)
   {
   ASCII_DB aDB (isIStream, ch_Separator, "");

   if (isIStream.fail ())
      {
      return (1);
      }

   while (aDB.Next () == 0)
      {
      if (InsertEntry (hashMap, aDB[0], aDB[1]))
         {
         cerr_Error << pcsz_ProgName << ": Can not insert data."
                    << " (" << pcszInFile << ")" << endl;
         return (1);
         }
      }

   return (0);
   }

/****************************************************************************/
/** Load a file into the HashMap.
***
*** File syntax is:
***   ...
***
*\C Parameters:
*\P   isIStream    IN - Stream to read from
*\P   pcszInFile   IN - Current stream name
*\P   hashMap     OUT - Where to put the data
***
*\C Returns:
*\R   0 == OK
*\N   1 == ERROR
*****************************************************************************/
APIRET LoadTable (istream & isIStream, PCSZ pcszInFile,
                  HASH_MAP_STR_STR & hashMap)
   {
   ASCII_DB aDB (isIStream, ch_Separator, "");
   ASCII_DB_RECORD adbHeader;
   ASCII_DB_RECORD adbData;

   if (isIStream.fail ())
      {
      return (1);
      }

   if ((aDB.Next () != 0) ||
       (aDB.CopyRecord (adbHeader) != 0) ||
       (aDB.Next () != 0))
      {
      cerr_Error << pcsz_ProgName << ": error readding Header "
                 << " (" << pcszInFile << ")" << endl;
      return (1);
      }

   while (aDB.Next () == 0)
      {
      if (aDB.CopyRecord (adbData) != 0)
         {
         cerr_Error << pcsz_ProgName << ": out of memory (?) "
                    << " (" << pcszInFile << ")" << endl;
         return (1);
         }
      }

   for (int i = 0; i < adbHeader.QueryFields (); i++)
      {
      if (InsertEntry (hashMap, adbHeader[i], adbData[i]))
         {
         cerr_Error << pcsz_ProgName << ": Can not insert data."
                    << " (" << pcszInFile << ")" << endl;
         return (1);
         }
      }

   return (0);
   }

