/****************************************************************************/
/* |\   ======   /| **              IDRepl              ** |\   ======   /| */
/* |__\========/__| **       MasarLabs ID-Replacer      ** |__\========/__| */
/* |\            /| ** -------------------------------- ** |\            /| */
/* |  \        /  | **  Maurizio Sartori (c) 1998-2000  ** |  \        /  | */
/* |    \    /    | ** -------------------------------- ** |    \    /    | */
/* |    /=\/=\    | **            Written  by           ** |    /=\/=\    | */
/* |  /========\  | **     Maurizio Sartori 'masar'     ** |  /========\  | */
/* |/   ======   \| **   e-mail:  masar@MasarLabs.com   ** |/   ======   \| */
/****************************************************************************/
/*              This program comes with ABSOLUTELY NO WARRANTY              */
/*            for details refer to the GNU General Public License           */
/****************************************************************************/
/*\F $Id: main.cpp,v 1.28 2003/11/19 15:47:31 cvs Exp $
*****************************************************************************/

static char Psz_RCS [] =
   "$Id: main.cpp,v 1.28 2003/11/19 15:47:31 cvs Exp $";

//****************************************************************************

#include <fstream>
#include <sstream>
#include <cstdlib>
#include <string>

#include "version.hpp"
#include "idrepl.hpp"
#include "verbose.hpp"

//****************************************************************************
// Globals
//****************************************************************************
#define IDREPL_DELIM_START     "$["
#define IDREPL_DELIM_STOP      "]"
#define IDREPL_ESCAPE          '%'
#define IDREPL_COMMENT         '#'
#define IDREPL_INCLUDE         '@'

PCSZ pcsz_ProgName   = NULL;
INT  i_MaxLevel      = 10;
BOOL f_Remove        = FALSE;
PCSZ pcsz_DelimStart = IDREPL_DELIM_START;
PCSZ pcsz_DelimStop  = IDREPL_DELIM_STOP;
CHAR ch_Separator    = '\t';
CHAR ch_Escape       = IDREPL_ESCAPE;
CHAR ch_Comment      = IDREPL_COMMENT;
CHAR ch_Include      = IDREPL_INCLUDE;

//****************************************************************************
//                                 main ()
//****************************************************************************
int main (int argc, const char * argv[])
   {
   *Psz_RCS = *Psz_RCS;                 // To fake compiler warnings

   APIRET apiRet = 0;
   ifstream ifDatFile;
   ifstream ifTemplate;
   ofstream ofOutFile;
   istream * isDat = &cin;
   istream * isTpl = &cin;
   ostream * osOut = &cout;
   HASH_MAP_STR_STR hashMap;
   HASH_MAP_STR_STR::iterator it;
   INT iFound;
   INT iReplaced;
   INT iError;
   INT i, j;
   STRING strFound;
   CHAR * pszStr;

   PCSZ pcszDatFile       = NULL;
   PCSZ pcszTemplate      = NULL;
   PCSZ pcszOutFile       = NULL;
   PCSZ pcszFormat        = "multi";
   INT  iMaxLevel         = 10;

   pcsz_ProgName = argv[0];
   MasarLabs_verbose = 0;
   //*************************************************************************
   // Parse Options
   //*************************************************************************
   i = 1;
   while (i < argc)
      {
      if (!strcmp (argv[i], "-c") || !strcmp (argv[i], "--comment"))
         {
         if ((i + 1) >= argc)
            {
            cerr_Error << pcsz_ProgName
                 << ": comment char is required" << endl;
            return (1);
            }
         ch_Comment = argv[++i][0];
         }
      else if (!strcmp (argv[i], "-d") || !strcmp (argv[i], "--delim"))
         {
         if ((i + 2) >= argc)
            {
            cerr_Error << pcsz_ProgName
                       << ": start and stop delimiters are required" << endl;
            return (1);
            }
         pcsz_DelimStart = argv[++i];
         pcsz_DelimStop  = argv[++i];
         }
      else if (!strcmp (argv[i], "-e") || !strcmp (argv[i], "--escape"))
         {
         if ((i + 1) >= argc)
            {
            cerr_Error << pcsz_ProgName
                 << ": escape char is required" << endl;
            return (1);
            }
         ch_Escape = argv[++i][0];
         }
      else if (!strcmp (argv[i], "-f") || !strcmp (argv[i], "--format"))
         {
         if ((i + 1) >= argc)
            {
            cerr_Error << pcsz_ProgName
                 << ": format string is required" << endl;
            return (1);
            }
         pcszFormat = argv[++i];
         }
      else if (!strcmp (argv[i], "-h") || !strcmp (argv[i], "--help"))
         {
         cout << "Syntax: " << endl
              << "   " << pcsz_ProgName << " [options]"
                                     " data template outfile" << endl
              << endl
              << "Options:" << endl
              << "   -c --comment <char>     comment char prefix (only for 'multi')"
                                             "(default: '" << IDREPL_COMMENT << "')" << endl
              << "   -d --delim str1 str2    use 'str1' and 'str2' as start"
                                             " and stop delimiters" << endl
              << "                           (default: '" IDREPL_DELIM_START
                                             "' '" IDREPL_DELIM_STOP "')" << endl
              << "   -e --escape <char>      escape char prefix "
                                             "(default: '" << IDREPL_ESCAPE << "')" << endl
              << "   -f --format <type>      set the input format: "
                                             "multi, list, table" << endl
              << "                           (default: '" << pcszFormat << "')" << endl
              << "   -h --help               print this help" << endl
              << "   -i --include <char>     include char prefix "
                                             "(default: '" << IDREPL_INCLUDE << "')" << endl
              << "   -m --maxlevel           maximum level of recursion "
                                             "(default: " << i_MaxLevel << ")"
                                             << endl
              << "   -q --quiet              decrease verbosity" << endl
              << "   -r --remove             remove undefined keys" << endl
              << "   -s --separator <char>   separator char for list and table "
                                             "(default: <tab>)" << endl
              << "   -v --verbose            increase verbosity" << endl
              << "   -V --version            print version" << endl
              ;

         return (0);
         }
      else if (!strcmp (argv[i], "-i") || !strcmp (argv[i], "--include"))
         {
         if ((i + 1) >= argc)
            {
            cerr_Error << pcsz_ProgName
                 << ": include char is required" << endl;
            return (1);
            }
         ch_Include = argv[++i][0];
         }
      else if (!strcmp (argv[i], "-m") || !strcmp (argv[i], "--maxlevel"))
         {
         ++i;
         if (i >= argc)
            {
            cerr_Error << pcsz_ProgName
                       << ": max level number is required" << endl;
            return (1);
            }
         i_MaxLevel = strtol (argv[i], &pszStr, 10);
         if ((pszStr == NULL) || (*pszStr) || (i_MaxLevel < 0))
            {
            cerr_Error << pcsz_ProgName
                       << ": invalid max level number: " << argv[i] << endl;
            return (1);
            }
         }
      else if (!strcmp (argv[i], "-q") || !strcmp (argv[i], "--quiet"))
         {
         MasarLabs_verbose--;
         }
      else if (!strcmp (argv[i], "-r") || !strcmp (argv[i], "--remove"))
         {
         f_Remove = TRUE;
         }
      else if (!strcmp (argv[i], "-s") || !strcmp (argv[i], "--separator"))
         {
         if ((i + 1) >= argc)
            {
            cerr_Error << pcsz_ProgName
                 << ": separator char is required" << endl;
            return (1);
            }
         ch_Separator = argv[++i][0];
         }
      else if (!strcmp (argv[i], "-v") || !strcmp (argv[i], "--verbose"))
         {
         MasarLabs_verbose++;
         }
      else if (!strcmp (argv[i], "-V") || !strcmp (argv[i], "--version"))
         {
         cout << pcsz_ProgName << endl
              << "  " APP_STR_NAME " Ver "
                 APP_STR_MAJOR "." APP_STR_MINOR "." APP_STR_REVIS " - "
                 APP_STR_DESCR "\n"
              << endl
              << "  Copyright " APP_STR_COPYRIGHT "\n"
                 "  Written by " APP_STR_AUTHOR " <" APP_STR_EMAIL ">\n"
              << endl
              << "  This program is free software; you may redistribute\n"
                 "  it under the terms of the GNU General Public License\n"
                 "  This program has absolutely no warranty\n"
              << endl
              << "  If you find this program useful, let me know,\n"
                 "  send a postcard with a photo of your city to:\n"
                 "\n"
                 "     Maurizio Sartori\n"
                 "     Via Caterina da Forli'  58\n"
                 "     20146 MILANO\n"
                 "     ITALY\n"
              << endl;
         return (0);
         }
      else if ((argv[i][0] == '-') && (argv[i][1] != 0))
         {
         cerr_Error << pcsz_ProgName << ": invalid parameter \""
                    << &argv[i][0] << "\"" << endl;
         return (1);
         }
      else
         {
         if (pcszDatFile == NULL)
            {
            pcszDatFile = argv[i];
            }
         else if (pcszTemplate == NULL)
            {
            pcszTemplate = argv[i];
            }
         else if (pcszOutFile == NULL)
            {
            pcszOutFile = argv[i];
            }
         else
            {
            cerr_Error << pcsz_ProgName << ": invalid argument \""
                       << argv[i] << "\"" << endl;
            return (1);
            }
         }
      ++i;
      }

   if (pcszDatFile == NULL)
      {
      cerr_Error << pcsz_ProgName
                 << ": data file is required (use -h for help)"
                 << endl;
      return (1);
      }

   if (pcszTemplate == NULL)
      {
      cerr_Error << pcsz_ProgName
                 << ": template file is required (use -h for help)"
                 << endl;
      return (1);
      }

   if (pcszOutFile == NULL)
      {
      cerr_Error << pcsz_ProgName
                 << ": output file is required (use -h for help)"
                 << endl;
      return (1);
      }

   //*************************************************************************
   // Initialize
   //*************************************************************************
   if (!strcmp (pcszDatFile, pcszTemplate))
      {
      cerr_Error << pcsz_ProgName
                 << ": data and template files cannot be the same"
                 << endl;
      return (1);
      }

   if (strcmp (pcszOutFile, "-") &&
       !strcmp (pcszOutFile, pcszTemplate))
      {
      cerr_Error << pcsz_ProgName
                 << ": output and template files cannot be the same"
                 << endl;
      return (1);
      }

   if (strcmp (pcszDatFile, "-") &&
       !strcmp (pcszDatFile, pcszOutFile))
      {
      cerr_Error << pcsz_ProgName
                 << ": data and output files cannot be the same"
                 << endl;
      return (1);
      }

   if (strcmp (pcszDatFile, "-"))
      {
      ifDatFile.open (pcszDatFile, ios::in);
      isDat = &ifDatFile;
      }

   if (isDat->fail ())
      {
      cerr_Error << pcsz_ProgName
                 << ": error opening input file" << endl;
      return (2);
      }

   if (strcmp (pcszTemplate, "-"))
      {
      ifTemplate.open (pcszTemplate, ios::in);
      isTpl = &ifTemplate;
      }

   if (isTpl->fail ())
      {
      cerr_Error << pcsz_ProgName
                 << ": error opening template file" << endl;
      return (3);
      }

   if (strcmp (pcszOutFile, "-"))
      {
      ofOutFile.open (pcszOutFile, ios::out);
      osOut = &ofOutFile;
      }

   if (osOut->fail ())
      {
      cerr_Error << pcsz_ProgName
                 << ": error opening output file" << endl;
      return (4);
      }

   //*************************************************************************
   // Read Data
   //*************************************************************************
   if (!strcmp (pcszFormat, "multi"))
      {
      if (LoadMulti (*isDat, pcszDatFile, hashMap))
         {
         return (1);
         }
      }
   else if (!strcmp (pcszFormat, "list"))
      {
      if (LoadList (*isDat, pcszDatFile, hashMap))
         {
         return (1);
         }
      }
   else if (!strcmp (pcszFormat, "table"))
      {
      if (LoadTable (*isDat, pcszDatFile, hashMap))
         {
         return (1);
         }
      }
   else
      {
      cerr_Error << pcsz_ProgName
                 << ": invalid format string \"" << pcszFormat
                 << "\"" << endl;
      return (1);
      }

   //*************************************************************************
   // Recurse
   //*************************************************************************
   j = 1;
   for (i = 0; (i < i_MaxLevel) && j; i++)
      {
      j = 0;

      for (it = hashMap.begin (); it != hashMap.end (); ++it)
         {
         istringstream isStrIn(it->second.c_str());
         ostringstream osStrOut;

         iReplaced = 0;
         iFound = 0;
         iError = 0;
         apiRet = IdReplace (isStrIn, osStrOut, hashMap,
                             iFound, iReplaced, iError,
                             pcsz_DelimStart, pcsz_DelimStop, f_Remove);
         if (iReplaced)
            {
            osStrOut << ends;
            it->second = osStrOut.str ();
//            osStrOut.freeze (0);
            j = 1;
            }
         }
      }

   iMaxLevel = i;

   if (j)
      {
      cerr_Error << pcsz_ProgName
                 << ": max level of recursion reached: "
                 << iMaxLevel << "."
                 << endl;
      return (4);
      }

   //*************************************************************************
   // Process
   //*************************************************************************
   for (it = hashMap.begin (); it != hashMap.end (); ++it)
      {
      cerr_DebugFull << "Key  = \"" << it->first << "\"\n"
                     << "Data = \"" << it->second << "\""
                     << endl;
      }

   cerr_Debug << "Delimiters '" << pcsz_DelimStart << "' and '"
              << pcsz_DelimStop << "'" << endl;

   iReplaced = 0;
   iFound = 0;
   iError = 0;
   apiRet = IdReplace (*isTpl, *osOut, hashMap,
                       iFound, iReplaced, iError,
                       pcsz_DelimStart, pcsz_DelimStop, f_Remove);

   if (strcmp (pcszDatFile, "-"))
      {
      ifDatFile.close ();
      }

   if (strcmp (pcszTemplate, "-"))
      {
      ifTemplate.close ();
      }

   if (strcmp (pcszOutFile, "-"))
      {
      ofOutFile.close ();
      }

   cerr_Debug << "Recursion " << iMaxLevel << endl
              << "Keys      " << hashMap.size() << endl
              << "Found     " << iFound << endl
              << "Replaced  " << iReplaced << endl
              << "Error     " << iError << endl
              << "Return    " << apiRet << endl;

   return ((apiRet == 0) ? 0 : 6);
   } /* End of main */

/****************************************************************************/
/** Insert an entry in the hash, performs recursive substitution.
***
*\C Parameters:
*\P   hashMap     I/O - Hash
*\P   strKey       IN - Key
*\P   strData      IN - Data
***
*\C Returns:
*\R   0 if OK
*****************************************************************************/
APIRET InsertEntry (HASH_MAP_STR_STR & hashMap,
                    STRING strKey, STRING strData)
   {
   APIRET apiRet = 0;
   INT iFound;
   INT iReplaced;
   INT iError;
   INT i, j;

   //*************************************************************************
   // Unescape
   //*************************************************************************
   if (UnEscape (strKey) || UnEscape (strData))
      {
      cerr_Error << pcsz_ProgName
                 << ": invalid escape sequence."
                 << endl;
      return (1);
      }

   /*************************************************************************/
   /* Recurse                                                               */
   /*************************************************************************/
   j = 1;
   for (i = 0; (i < i_MaxLevel) && j; i++)
      {
      istringstream isStrIn(strData.c_str());
      ostringstream osStrOut;
      j = 0;

      iReplaced = 0;
      iFound = 0;
      iError = 0;
      apiRet = IdReplace (isStrIn, osStrOut, hashMap,
                          iFound, iReplaced, iError,
                          pcsz_DelimStart, pcsz_DelimStop, FALSE);

      if (!apiRet && iReplaced)
         {
         osStrOut << ends;
         strData = osStrOut.str ();
//         osStrOut.freeze (0);
         j = 1;
         }
      }

   if (j)
      {
      cerr_Error << pcsz_ProgName
                 << ": max level of recursion reached: "
                 << j << "."
                 << endl;
      return (4);
      }

   hashMap[strKey] = strData;

   return (0);
   }

/****************************************************************************/
/** Unescape a string.
*** An escaped char has the form '%XX' where 'X' is an hexadecimal value.
*** Some usefull escapes:
***  '%' -> '%25'
***  ' ' -> '%20'
***  NL  -> '%0A'
***  CR  -> '%0D'
***
*\C Parameters:
*\P   strData     I/O - String to unescape
***
*\C Globals:
*\P   ch_Escape    IN - Escape char
***
*\C Returns:
*\R   0 if OK
*****************************************************************************/
APIRET UnEscape (STRING & strData)
   {
   STRING strTmp;
   STRING::const_iterator it;
   UCHAR uch;

   for (it = strData.begin (); it != strData.end (); ++it)
      {
      if (*it == ch_Escape)
         {
         uch = 0;
         ++it;
         if (it == strData.end ())
            {
            return (1);
            }

         if ((*it >= '0') && (*it <= '9'))
            {
            uch += (*it - '0') * 16;
            }
         else if ((*it >= 'a') && (*it <= 'f'))
            {
            uch += (*it - 'a' + 10) * 16;
            }
         else if ((*it >= 'A') && (*it <= 'F'))
            {
            uch += (*it - 'A' + 10) * 16;
            }
         else
            {
            return (2);
            }

         ++it;
         if (it == strData.end ())
            {
            return (1);
            }

         if ((*it >= '0') && (*it <= '9'))
            {
            uch += (*it - '0');
            }
         else if ((*it >= 'a') && (*it <= 'f'))
            {
            uch += (*it - 'a' + 10);
            }
         else if ((*it >= 'A') && (*it <= 'F'))
            {
            uch += (*it - 'A' + 10);
            }
         else
            {
            return (2);
            }
         strTmp += uch;
         }
      else
         {
         strTmp += *it;
         }
      }

   strData = strTmp;
   return (0);
   }

