/****************************************************************************/
/* |\   ======   /| **              IDRepl              ** |\   ======   /| */
/* |__\========/__| **       MasarLabs ID-Replacer      ** |__\========/__| */
/* |\            /| ** -------------------------------- ** |\            /| */
/* |  \        /  | **  Maurizio Sartori (c) 1998-2000  ** |  \        /  | */
/* |    \    /    | ** -------------------------------- ** |    \    /    | */
/* |    /=\/=\    | **            Written  by           ** |    /=\/=\    | */
/* |  /========\  | **     Maurizio Sartori 'masar'     ** |  /========\  | */
/* |/   ======   \| **   e-mail:  masar@MasarLabs.com   ** |/   ======   \| */
/****************************************************************************/
/*              This program comes with ABSOLUTELY NO WARRANTY              */
/*            for details refer to the GNU General Public License           */
/****************************************************************************/
/*\F $Id: multi.cpp,v 1.18 2003/11/19 16:55:33 cvs Exp $
*****************************************************************************/

static char Psz_RCS [] =
   "$Id: multi.cpp,v 1.18 2003/11/19 16:55:33 cvs Exp $";

//****************************************************************************

#include <fstream>
#include <cerrno>
#include "idrepl.hpp"
#include "verbose.hpp"

/****************************************************************************/
/** Print a string on standard error.
***
*\C Parameters:
*\P   pcszMessage  IN - Message
*\P   pcszInFile   IN - Current stream name
*\P   iCol         IN - Column position
*\P   iLine        IN - Line position
***
*\C Returns:
*\R   0 if OK
*****************************************************************************/
static inline VOID WriteError (PCSZ pcszMessage, PCSZ pcszInFile,
                               INT iCol, INT iLine)
   {
   cerr_Error << pcsz_ProgName << ": " << pcszMessage
              << " (" << pcszInFile << " line: " << iLine
              << " col: " << iCol << ")"
              << endl;
   }

/****************************************************************************/
/** Read a char from a stream.
***
*\C Parameters:
*\P   isIStream    IN - Stream to read from
*\P   chData      OUT - Char read
*\P   iCol        I/O - Column position
*\P   iLine       I/O - Line position
***
*\C Returns:
*\R   0 if OK
*****************************************************************************/
static inline APIRET ReadChar (istream & isIStream, char & chData,
                               INT & iCol, INT & iLine)
   {
   chData = 0;

   if (isIStream.fail ())
      {
      return (1);
      }

   isIStream.get (chData);
   if (isIStream.fail ())
      {
      return (1);
      }

   ++iCol;
   if (chData == '\n')
      {
      ++iLine;
      iCol = 1;
      }

   return (0);
   }


/****************************************************************************/
/** Load a file into the HashMap.
***
*** File syntax is:
***   Comment starts with # in first Position.
***   Empty lines are ignored (lines with only spaces).
***   Data starts with a non space in first position:
***   the key must not contain spaces, data may split more lines
***   if the first char in a line is a space.
***
*\C Parameters:
*\P   isIStream    IN - Stream to read from
*\P   pcszInFile   IN - Current stream name
*\P   hashMap     OUT - Where to put the data
***
*\C Globals:
*\P   ch_Comment   IN - Comment char prefix
***
*\C Returns:
*\R   0 if OK
*****************************************************************************/
APIRET LoadMulti (istream & isIStream, PCSZ pcszInFile,
                  HASH_MAP_STR_STR & hashMap)
   {
   *Psz_RCS = *Psz_RCS;                 // To fake compiler warnings

   CHAR  chData;
   STRING strKey;
   STRING strData;
   STRING strStrip;
   INT iCol;
   INT iLine;

   if (isIStream.fail ())
      {
      return (1);
      }

   //*************************************************************************
   // Search
   //*************************************************************************
   iCol = 1;
   iLine = 1;

   strKey   = "";
   strData  = "";
   strStrip = "";

   while (ReadChar (isIStream, chData, iCol, iLine) == 0)
      {
      if (chData == '\n')
         {
         //*******************************************************************
         // Empty
         //*******************************************************************
         // NOP
         }
      else if (isspace (chData))
         {
         //*******************************************************************
         // Syntax error
         //*******************************************************************
         while ((ReadChar (isIStream, chData, iCol, iLine) == 0) &&
                (chData != '\n') && (isspace (chData)))
            {
            // NOP
            }
         if (!(isIStream.fail ()) && (chData != '\n'))
            {
            WriteError ("Not empty line and not data!",
                        pcszInFile, iCol, iLine);
            return (1);
            }
         }
      else if (chData == ch_Comment)
         {
         //*******************************************************************
         // Comment
         //*******************************************************************
         while ((ReadChar (isIStream, chData, iCol, iLine) == 0) &&
                (chData != '\n'))
            {
            // NOP
            }
         if (isIStream.fail ())
            {
            WriteError ("Not final newline! (1)",
                        pcszInFile, iCol, iLine);
            return (1);
            }
         }
      else if (chData == ch_Include)
         {
         //*******************************************************************
         // Include
         //*******************************************************************
         strKey   = "";

         while ((ReadChar (isIStream, chData, iCol, iLine) == 0) &&
                (chData != '\n'))
            {
            strKey += chData;
            }

         if (isIStream.fail ())
            {
            WriteError ("Not final newline! (2)",
                        pcszInFile, iCol, iLine);
            return (1);
            }

         ifstream ifInFile;

         ifInFile.open (strKey.c_str(), ios::in);
         if (ifInFile.fail ())
            {
            CHAR achTmp[20];
            sprintf (achTmp, "%i", errno);
            STRING strTmp = "Error opening input file \"";
            strTmp += strKey;
            strTmp += "\" (errno = ";
            strTmp += achTmp;
            strTmp += ")";

            WriteError (strTmp.c_str (),
                        pcszInFile, iCol, iLine);
            return (1);
            }

         if (LoadMulti (ifInFile, pcszInFile, hashMap))
            {
            return (1);
            }

         ifInFile.close ();
         }
      else
         {
         //*******************************************************************
         // Key
         //*******************************************************************
         strKey   = "";
         strData  = "";
         strStrip = "";

         strKey += chData;
         while ((ReadChar (isIStream, chData, iCol, iLine) == 0) &&
                !isspace (chData))
            {
            strKey += chData;
            }

         if (isIStream.fail ())
            {
            WriteError ("Not final newline! (3)",
                        pcszInFile, iCol, iLine);
            return (1);
            }
         isIStream.unget ();

         //*******************************************************************
         // Data
         //*******************************************************************
         while (!isIStream.fail ())
            {
            if (ReadChar (isIStream, chData, iCol, iLine))
               {
               break;
               }

            if (chData == '\n')
               {
               continue;
               }

            if (!isspace (chData))
               {
               isIStream.unget ();
               break;
               }

            while ((ReadChar (isIStream, chData, iCol, iLine) == 0) &&
                   isspace (chData) && (chData != '\n'))
               {
               // Skipp Spaces
               }

            if (isIStream.fail ())
               {
               WriteError ("Not final newline! (5)",
                           pcszInFile, iCol, iLine);
               return (1);
               }

            if (chData == '\n')
               {
               continue;
               }

            if (chData == '.')
               {
               if (ReadChar (isIStream, chData, iCol, iLine))
                  {
                  WriteError ("Not final newline! (6)",
                              pcszInFile, iCol, iLine);
                  return (1);
                  }

               if (chData == '\n')
                  {
                  continue;
                  }
               }

            strStrip = "";
            do
               {
               if (isspace (chData))
                  {
                  strStrip += chData;
                  }
               else
                  {
                  if (strStrip.length ())
                     {
                     strData += strStrip;
                     strStrip = "";
                     }
                  strData += chData;
                  }
               }
            while ((ReadChar (isIStream, chData, iCol, iLine) == 0) &&
                   (chData != '\n'));
            }

         if (InsertEntry (hashMap, strKey, strData))
            {
            WriteError ("Can not insert data.",
                        pcszInFile, iCol, iLine);
            return (1);
            }
         }
      }

   return (0);
   }

