/****************************************************************************/
/* |\   ======   /| **              IDRepl              ** |\   ======   /| */
/* |__\========/__| **       MasarLabs ID-Replacer      ** |__\========/__| */
/* |\            /| ** -------------------------------- ** |\            /| */
/* |  \        /  | **  Maurizio Sartori (c) 1998-2000  ** |  \        /  | */
/* |    \    /    | ** -------------------------------- ** |    \    /    | */
/* |    /=\/=\    | **            Written  by           ** |    /=\/=\    | */
/* |  /========\  | **     Maurizio Sartori 'masar'     ** |  /========\  | */
/* |/   ======   \| **   e-mail:  masar@MasarLabs.com   ** |/   ======   \| */
/****************************************************************************/
/*              This program comes with ABSOLUTELY NO WARRANTY              */
/*            for details refer to the GNU General Public License           */
/****************************************************************************/
/*\F $Id: replace.cpp,v 1.9 2000/12/17 19:18:04 cvs Exp $
*****************************************************************************/

static char Psz_RCS [] =
   "$Id: replace.cpp,v 1.9 2000/12/17 19:18:04 cvs Exp $";

//****************************************************************************

#include <fstream>
#include "idrepl.hpp"
#include "verbose.hpp"

/****************************************************************************/
/** Replace tags.
***
*\C Parameters:
*\P   isIStream   IN - Stream to read template file
*\P   osOStream  OUT - Stream to write output data file
*\P   hashMap     IN - Keys to be replaced
*\P   iFound     OUT - Number of keys found in template file
*\P   iReplaced  OUT - Number of keys replaced in template file
*\P   iError     OUT - Number of errors found
*\P   pcszPrefix  IN - Key prefix string
*\P   pcszSuffix  IN - Key suffix string
*\P   fRemove     IN - Remove not found keys
***
*\C Returns:
*\R   0 == OK
*\N   1 == ERROR
*****************************************************************************/
APIRET IdReplace (istream & isIStream, ostream & osOStream,
                  const HASH_MAP_STR_STR & hashMap,
                  INT &iFound, INT & iReplaced, INT &iError,
                  PCSZ pcszPrefix, PCSZ pcszSuffix, BOOL fRemove)
   {
   HASH_MAP_STR_STR::const_iterator itMap;
   ssize_t iSize;
   ssize_t iMaxKeySize;
   STRING strKey;
   ifstream ifInFile;
   APIRET iApiRet = 0;

   *Psz_RCS = *Psz_RCS;                 // To fake compiler warnings

   //*************************************************************************
   // Find Max Id Size
   //*************************************************************************
   iMaxKeySize = 0;
   if (fRemove)
      {
      iMaxKeySize = 0;
      itMap = hashMap.begin ();
      while (itMap != hashMap.end ())
         {
         iSize = itMap->first.length ();
         if (iSize > iMaxKeySize)
            {
            iMaxKeySize = iSize;
            }
         ++itMap;
         }
      ++iMaxKeySize;
      }
   else
      {
      iMaxKeySize = -1;
      }

   //*************************************************************************
   // Process
   //*************************************************************************
   iFound = 0;
   iReplaced = 0;
   iError = 0;
   while (StrSearchSkip (pcszPrefix, isIStream, osOStream))
      {
      if (StrSearchSkip (pcszSuffix, isIStream, strKey, iMaxKeySize))
         {
         ++iFound;
         itMap = hashMap.find (strKey);
         if (itMap != hashMap.end ())
            {
            if (itMap->second[0] != ch_Include)
               {
               osOStream.write (itMap->second.c_str (),
                                itMap->second.length ());
               }
            else
               {
               ifInFile.open ((itMap->second.c_str())+1, ios::in);
               if (ifInFile.fail ())
                  {
                  cerr_Error << pcsz_ProgName
                             << ": cannot open file \""
                             << (itMap->second.c_str())+1
                             << "\""
                             << endl;
                  iApiRet = 1;
                  }
               else
                  {
                  osOStream << ifInFile.rdbuf ();
                  ifInFile.close ();
                  }
               }

            ++iReplaced;
            }
         else if (!fRemove)
            {
            osOStream << pcszPrefix << strKey << pcszSuffix;
            }
         else
            {
            ++iReplaced;
            }
         }
      else
         {
         cerr_Error << pcsz_ProgName
                    << ": non terminated tag ??"
                    << endl;
         }
      }

   return (iApiRet);
   }

/****************************************************************************/
/** Search the pattern in a stream and put data read to strOut,
*** stops to read when pattern is found.
*** The pattern is not written to strOut.
***
*\C Parameters:
*\P   pcszPattern  IN - String to search
*\P   isIStream    IN - Stream to read from
*\P   strOut       IN - String to write to
*\P   iMaxLen      IN - Put not more than iMaxLen chars in strOut, discard
***                     the remaining (-1 == Unlimited)
***
*\C Returns:
*\R   FALSE if not found
*\R   TRUE  if found
*****************************************************************************/
BOOL StrSearchSkip (PCSZ pcszPattern, istream & isIStream,
                    STRING & strOut, INT iMaxLen)
   {
   *Psz_RCS = *Psz_RCS;                 // To fake compiler warnings

   INT   iPattLen;
   INT * piNext;
   INT   i, j, k;
   CHAR  chData;

   if ((pcszPattern == NULL) || isIStream.fail ())
      {
      return (FALSE);
      }

   strOut = "";
   iPattLen = strlen (pcszPattern);

   piNext = new INT[iPattLen+1];

   if (piNext == NULL)
      {
      return (FALSE);
      }

   //*************************************************************************
   // Init piNext
   //*************************************************************************
   piNext [0] = -1;
   for (i = 0, j = -1; (i < iPattLen); ++i, ++j, piNext[i] = j)
      {
      while ((j >= 0) && (pcszPattern [i] != pcszPattern [j]))
         {
         j = piNext [j];
         }
      }

   //*************************************************************************
   // Search
   //*************************************************************************
   isIStream.get (chData);
   for (j = 0; (j < iPattLen) && !isIStream.fail (); ++j)
      {
      while ((j >= 0) && (chData != pcszPattern[j]))
         {
         if (piNext[j] != -1)
            {
            for (k = 0; (k < j - piNext[j]) && (iMaxLen != 0); ++k)
               {
               strOut += pcszPattern[k];
               if (iMaxLen > 0)
                  {
                  --iMaxLen;
                  }
               }
            }
         j = piNext[j];
         }

      if ((j == -1) && (iMaxLen != 0))
         {
         strOut += chData;
         }
      isIStream.get (chData);
      }

   if (!isIStream.fail ())
      {
      isIStream.putback (chData);
      }

   delete [] piNext;

   return (j == iPattLen);
   }

/****************************************************************************/
/** Search the pattern in a stream and put data read to osOStream,
*** stops to read when pattern is found.
*** The pattern is not written to osOStream.
***
*\C Parameters:
*\P   pcszPattern  IN - String to search
*\P   isIStream    IN - Stream to read from
*\P   osOStream    IN - Stream to write to
***
*\C Returns:
*\R   FALSE if not found
*\R   TRUE  if found
*****************************************************************************/
BOOL StrSearchSkip (PCSZ pcszPattern,
                    istream & isIStream, ostream & osOStream)
   {
   *Psz_RCS = *Psz_RCS;                 // To fake compiler warnings

   INT   iPattLen;
   INT * piNext;
   INT   i, j, k;
   CHAR  chData;

   if ((pcszPattern == NULL) || isIStream.fail ())
      {
      return (FALSE);
      }

   iPattLen = strlen (pcszPattern);

   piNext = new INT[iPattLen+1];

   if (piNext == NULL)
      {
      return (FALSE);
      }

   //*************************************************************************
   // Init piNext
   //*************************************************************************
   piNext [0] = -1;
   for (i = 0, j = -1; (i < iPattLen); ++i, ++j, piNext[i] = j)
      {
      while ((j >= 0) && (pcszPattern [i] != pcszPattern [j]))
         {
         j = piNext [j];
         }
      }

   //*************************************************************************
   // Search
   //*************************************************************************
   isIStream.get (chData);
   for (j = 0; (j < iPattLen) && !isIStream.fail (); ++j)
      {
      while ((j >= 0) && (chData != pcszPattern[j]))
         {
         if (piNext[j] != -1)
            {
            for (k = 0; k < j - piNext[j]; ++k)
               {
               osOStream.put (pcszPattern[k]);
               }
            }
         j = piNext[j];
         }

      if (j == -1)
         {
         osOStream.put (chData);
         }
      isIStream.get (chData);
      }

   if (!isIStream.fail ())
      {
      isIStream.putback (chData);
      }

   delete [] piNext;

   return (j == iPattLen);
   }

